import logging
import traceback
import threading
import collections
import serial
from serial import Serial, SerialException
from aci.aci_cmd import response_deserialize
from aci.aci_cmd import CommandPacket
from aci.aci_cmd import ResponsePacket
from aci.aci_utils import EventPacket
from aci.aci_cmd import CmdRsp
from aci.aci_cmd import RESPONSE_LUT
from aci.aci_utils import prettify_data
from aci.aci_evt import event_deserialize
from aci.aci_utils import value_to_barray
import aci.aci_cmd as cmd
from aci.aci_utils import STATUS_CODE_LUT
from aci.aci_utils import iterable_to_barray
import aci.aci_evt as evt
import queue
import atexit
from queue import Queue, Empty

EVT_Q_BUF = 24
SEGGER_UART_BYTES_MAX = 24

# Configure logging
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

class Device(object):
    def __init__(self, device_name):
        self.device_name = device_name
        self.logger = logging.getLogger(self.device_name)
        self._pack_recipients = []
        self._cmd_recipients = []
        #self.lock = threading.Event()
        self.lock = threading.Lock() 
        self.condition = threading.Condition(self.lock)  # Use Condition with the lock
        self.events = list()
        self.__write_queue = queue.Queue()
        self.writer_alive = True
        self.writer_thread = threading.Thread(target=self.__writer, daemon=True)
        self.writer_thread.start()
        self.logger.info("Device initialized: %s", device_name)

    def __del__(self):
        self.kill_writer()

    def kill_writer(self):
        self.writer_alive = False
        self.__write_queue.put(None)
        if self.writer_thread.is_alive():
            self.writer_thread.join()
        self.logger.info("Writer thread killed")

    def __wait(self, timeout=2):
        self.logger.info("Waiting for event...")
        with self.lock:  # Acquire the lock
            self.logger.info("Waiting for event with timeout %s...", timeout)
            self.condition.wait(timeout)
            if len(self.events) == 0:
                self.logger.info("Timeout waiting for event.")
                return None
            else:
                event = self.events[:]
                self.events.clear()
                self.logger.info("Returning event: %s", event)
                return event


    def add_packet_recipient(self, function):
        self._pack_recipients.append(function)

    def remove_packet_recipient(self, function):
        if function in self._pack_recipients:
            self._pack_recipients.remove(function)

    def add_command_recipient(self, function):
        self._cmd_recipients.append(function)

    def process_packet(self, packet):
        self.logger.info("Processing packet...")
        with self.lock:  # Acquire the lock
            self.events.append(packet)
            self.logger.info("Packet processed, event set.")
        for fun in self._pack_recipients[:]:
            try:
                self.logger.info("DAMN RECURSIVE.")
                fun(packet)
            except Exception as e:
                self.logger.error('Exception in pkt handler %r: %s', fun, e)
                self.logger.error('Traceback: %s', traceback.format_exc())



    def process_command(self, command):
        for fun in self._cmd_recipients[:]:
            try:
                fun(command)
            except Exception as e:
                self.logger.error('Exception in cmd handler %r: %s', fun, e)
                self.logger.error('Traceback: %s', traceback.format_exc())

    def __writer(self):
        while self.writer_alive:
            try:
                self.logger.info("Waiting for command...")
                cmd = self.__write_queue.get()
                self.logger.info("Command dequeued: %s", cmd)
                if cmd is None:
                    self.logger.info("Stopping writer thread.")
                    return

                self.logger.info("Writing command: %s", cmd)
                cmd.logger = self.logger
                
                # Ensure the lock is acquired before writing data
                with self.lock:
                    self.write_data(cmd)  # Directly call the method to write to serial

                self.logger.info("Command written, waiting for response...")
                retval = self.__wait()
                if retval is None:
                    self.logger.info('Command %s: Timeout waiting for event', cmd.__class__.__name__)
                else:
                    self.logger.info('Command %s: Event received', cmd.__class__.__name__)

            except Exception as e:
                self.logger.error("Exception in thread %s: %s", threading.current_thread().name, e)


    def write_aci_cmd(self, cmd):
        self.__write_queue.put(cmd)


    def write_data(self, data):
        self.logger.error("write_data method must be implemented by subclasses.")



class Uart(threading.Thread, Device):
    def __init__(self, port, baudrate=115200, device_name=None, rtscts=False):
        #super().__init__(device_name or port)
        self.events_queue = collections.deque(maxlen=EVT_Q_BUF)
        threading.Thread.__init__(self)
        self.logger = logging.getLogger('Uart')
        self.port = port
        self.baudrate = baudrate
        self.max_retries = 3
        self.rtscts = rtscts
        self.keep_running = True
        Device.__init__(self, device_name or port)
        self._write_lock = threading.Lock()
        self.packet_recipients = []
        self.command_queue = Queue()
        self.serial = None

        try:
            self.serial = Serial(port=port, baudrate=baudrate, rtscts=rtscts, timeout=1)
            self.logger.info("UART initialized on port %s", self.port)
        except SerialException as e:
            self.logger.error("Failed to initialize UART on port %s: %s", self.port, e)
            self.serial = None

        self.read_thread = threading.Thread(target=self._read_from_port)
        self.read_thread.daemon = True
        self.read_thread.start()

        self.write_thread = threading.Thread(target=self._process_commands)
        self.write_thread.daemon = True
        self.write_thread.start()

    def add_0x6D_to_first_byte(self, byte_array):
        """
        Add 0x6D to the first byte in the given bytearray. If the resulting value exceeds 256,
        the original byte value is retained.

        Parameters:
        byte_array (bytearray): The input bytearray.

        Returns:
        bytearray: The bytearray with 0x6D added to the first byte.
        """
        result = bytearray()
        for index, byte in enumerate(byte_array):
            if index == 0:
                new_byte = byte + 0x6D
                if new_byte <= 0xFF:
                    result.append(new_byte)
                else:
                    result.append(byte)
            else:
                result.append(byte)
        return result


    def reverse_bits_in_byte(self, byte):
        if not isinstance(byte, int) or not 0 <= byte <= 0xFF:
            raise ValueError("Input must be a single byte (0-255).")

        reversed_byte = 0
        for i in range(8):
            reversed_byte <<= 1
            reversed_byte |= (byte >> i) & 1
        return reversed_byte

    def reverse_bytes_in_array(self, byte_array):
        if not isinstance(byte_array, bytearray):
            raise TypeError("Input must be a bytearray.")

        reversed_array = bytearray()
        for byte in byte_array:
            reversed_byte = self.reverse_bits_in_byte(byte)
            reversed_array.append(reversed_byte)

        return reversed_array


    def _read_from_port(self):
        while self.keep_running and self.serial and self.serial.is_open:
            try:
                data = self.serial.read(self.serial.in_waiting or 1)
                converted_data = bytearray(data)
                # Uncomment and use if needed: self.add_0x6D_to_first_byte(converted_data)

                if converted_data and len(converted_data) >= 2:
                    self.logger.debug("RX: %s", converted_data.hex())
                    #op = converted_data.hex()[0:2]
                    # Deserialize event
            
                    event = event_deserialize(converted_data.hex())
                        
                    if event:
                        self.events_queue.append(event)
                        self.process_packet(event)
                        print(event)
                    else:
                        cmd_des = cmd.response_deserialize(converted_data.hex())
                        if cmd_des and isinstance(cmd_des, CmdRsp):
                            self.write_aci_cmd(cmd_des)
                            self.logger.debug("Deserialized COMMAND is: %s", cmd_des)
                            self.logger.info('cmd %s, timeout waiting for event' % (cmd_des.__class__.__name__))
                        #command_er = response_deserialize(converted_data)
                        #if isinstance(command_er, CmdRsp):
                            #if command_er:
                                #self.write_data(data)
                                #if not isinstance(command_er, ResponsePacket):
                               # self.write_aci_cmd(command_er)
                              #  self.logger.debug("Deserialized COMMAND is: %s", command_er)
                             #   self.logger.info('cmd %s, timeout waiting for event' % (command_er.__class__.__name__))
                            #    self.process_command(command_er)
                            #else:
                                #self.logger.debug("Deserialized COMMAND is None, data: %s", converted_data.hex())
                        #if command_er:
                            #self.events_queue.append(command_er)
                            #self.write_data(command_er)
                            #self.process_command(command_er)
                        #if not response_deserialize_sec(data):
                         #   response_deserialize_sec(self.reverse_bytes_in_array(bytearray(data)))
                        

            except SerialException as e:
                self.logger.error("Error reading from serial port: %s", e)
                self.keep_running = False



    def write_data(self, data):
        self.command_queue.put(data)

    def _process_commands(self):
        while self.keep_running:
            try:
                data = self.command_queue.get(timeout=1)  # Wait for a command
                self._write_command(data)
                self.command_queue.task_done()
            except Empty:
                continue  # No command to process

    def _write_command(self, data):
        if self.serial and self.serial.is_open:
            serialized_data = data.serialize()  # Ensure this method exists or format data correctly
            with self._write_lock:
                if self.keep_running:
                    self.logger.debug("TX: %s", bytearray(serialized_data).hex())
                    while len(serialized_data) > 0:
                        try:
                            self.serial.write(bytearray(serialized_data[:SEGGER_UART_BYTES_MAX]))
                            serialized_data = serialized_data[SEGGER_UART_BYTES_MAX:]
                        except serial.SerialTimeoutException:
                            self.logger.warning("Write timeout ignored.")
                        except serial.SerialException as e:
                            self.logger.error("SerialException occurred: %s", e)
                            raise
                    self.process_command(data)
                    self.logger.info("Data written to UART: %s", data)
        else:
            self.logger.error("Serial port not open. Cannot write data.")

    def get_packet_from_uart(self):
        tmp = bytearray([])
        while self.keep_running:
            tmp += bytearray(self.serial.read())
            tmp_len = len(tmp)
            if tmp_len > 0:
                pkt_len = tmp[0]
                if tmp_len > pkt_len:
                    data = tmp[:pkt_len+1]
                    yield data
                    tmp = tmp[pkt_len+1:]
    
    def close(self):
        # Assuming close method should perform the necessary cleanup
        # If close method exists and does the required cleanup
        pass

    # Define the close method if it's missing or adjust if necessary
    def stop(self):
        self.close()  # Redirect stop to close

    #def run(self):
        #for pkt in self.get_packet_from_uart():
           # self.logger.debug("RX: %s", pkt.hex())
           # try:
            #    if len(pkt) < 2:
            #        self.logger.error('Invalid packet: %r', pkt)
             #       continue
             #   parsed_packet = event_deserialize(pkt)
             #   if not parsed_packet:
             #       self.logger.error("Unable to deserialize %s", pkt.hex())

            #except Exception:
            #    self.logger.error('Exception with packet %s', pkt.hex())
            #    self.logger.error('traceback: %s', traceback.format_exc())
            #    parsed_packet = None

           # if parsed_packet:
           #     self.events_queue.append(parsed_packet)
          #      self.logger.debug('parsed_packet %r', parsed_packet)
         #       self.process_packet(parsed_packet)

        #self.serial.close()
        #self.logger.debug("exited read event")

    #def stop(self):
     #   self.serial.close()


    def close(self):
        self.keep_running = False
        if self.serial and self.serial.is_open:
            self.serial.close()
        self.logger.info("UART connection closed.")

